@@ -1,11 +1,15 @@
 use std::{fmt::Display, sync::Arc};
 
+use arrow::ffi_stream::ArrowArrayStreamReader;
+use arrow::record_batch::RecordBatchReader;
 use arrow::{
     datatypes::Schema, error::ArrowError, ipc::reader::StreamReader, record_batch::RecordBatch,
 };
 use futures::stream::TryStreamExt;
 use log::info;
 use tiberius::Client;
+use tiberius::ColumnType;
+use tiberius::SqlBulkCopyOptions;
 use tokio::net::TcpStream;
 use tokio_util::compat::Compat;
 use tokio_util::compat::FuturesAsyncReadCompatExt;
@@ -61,9 +65,49 @@ impl std::error::Error for SendErrorWrap {
     }
 }
 
+async fn get_cols_from_table(
+    db_client: &mut Client<Compat<TcpStream>>,
+    table_name: &str,
+    column_names: &[&str],
+) -> Result<Vec<(String, ColumnType)>, Box<dyn std::error::Error + Send + Sync>> {
+    let cols_sql = match column_names.len() {
+        0 => "*".to_owned(),
+        _ => column_names
+            .iter()
+            .map(|c| format!("[{}]", c))
+            .collect::<Vec<String>>()
+            .join(", "),
+    };
+    let query = format!("SELECT TOP 0 {} FROM {}", cols_sql, table_name);
+    let mut colres = db_client.simple_query(query).await?;
+    Ok(colres
+        .columns()
+        .await?
+        .unwrap()
+        .iter()
+        .map(|x| (x.name().to_string(), x.column_type()))
+        .collect::<Vec<(String, ColumnType)>>())
+}
+
+pub async fn bulk_insert_batch<'a>(
+    blk: &mut tiberius::BulkLoadRequest<'a, Compat<TcpStream>>,
+    batch: &'a RecordBatch,
+    collist: &'a Vec<(String, ColumnType)>,
+) -> Result<(), Box<dyn std::error::Error + Send + Sync>> {
+    let nrows = batch.num_rows();
+    info!("received {nrows}");
+    let rows = get_token_rows(batch, &collist)?;
+    for rowdt in rows {
+        blk.send(rowdt).await?;
+    }
+    info!("Written {nrows}");
+    Ok(())
+}
+
 pub async fn bulk_insert<'a>(
     db_client: &'a mut Client<Compat<TcpStream>>,
     table_name: &str,
+    column_names: &'a [&'a str],
     url: &str,
     user: &str,
     password: &str,
@@ -72,29 +116,9 @@ pub async fn bulk_insert<'a>(
     //row.push(1.into_sql());
     //blk.send(row).await?;
     //blk.finalize().await?;
-    let colres = db_client
-        .query(
-            "
-    select sc.name from sys.columns sc
-        inner join sys.objects o on o.object_id=sc.object_id
-        inner join sys.schemas sch on sch.schema_id=o.schema_id  
-        where is_computed=0 and is_identity=0 and system_type_id<>189
-            and o.name=parsename(@P1,1)
-            and sch.name=coalesce(parsename(@P1,2), SCHEMA_NAME())
-        order by sc.column_id
-    ",
-            &[&table_name],
-        )
-        .await?;
-    let res = colres.into_first_result().await?;
-    let collist = res
-        .iter()
-        .map(|row| {
-            let val: &str = row.get(0).unwrap();
-            return val.to_string();
-        })
-        .collect::<Vec<String>>();
 
+    let collist = get_cols_from_table(db_client, table_name, column_names).await?;
+    log::debug!("{:?}", collist);
     let cclient = reqwest::Client::new();
 
     // a bit too complex if you ask me: https://github.com/benkay86/async-applied/tree/master/reqwest-tokio-compat
@@ -140,19 +164,55 @@ pub async fn bulk_insert<'a>(
             Ok(schema)
         },
     );
-
     while let Some(v) = rx.recv().await {
-        let nrows = v.num_rows();
-        info!("received {nrows}");
-        let rows = get_token_rows(&v, &collist)?;
-        let mut blk: tiberius::BulkLoadRequest<'_, Compat<TcpStream>> =
-            db_client.bulk_insert(table_name).await?;
-        for row in rows {
-            blk.send(row).await?;
-        }
+        let mut blk = db_client
+            .bulk_insert_with_options(
+                table_name,
+                &column_names,
+                SqlBulkCopyOptions::TableLock,
+                &[],
+            )
+            .await?;
+        bulk_insert_batch(&mut blk, &v, &collist).await?;
         blk.finalize().await?;
-        info!("Written {nrows}");
     }
     let schema = worker.await?;
     schema
 }
+
+pub async fn bulk_insert_reader(
+    db_client: &mut Client<Compat<TcpStream>>,
+    table_name: &str,
+    column_names: &[&str],
+    reader: &mut ArrowArrayStreamReader,
+) -> Result<Arc<Schema>, Box<dyn std::error::Error + Send + Sync>> {
+    //let mut row = TokenRow::new();
+    //row.push(1.into_sql());
+    //blk.send(row).await?;
+    //blk.finalize().await?;
+
+    let collist = get_cols_from_table(db_client, table_name, column_names).await?;
+    log::debug!("{:?}", collist);
+    let schema = reader.schema();
+    loop {
+        match reader.next() {
+            Some(x) => match x {
+                Ok(b) => {
+                    let mut blk = db_client
+                        .bulk_insert_with_options(
+                            table_name,
+                            &column_names,
+                            SqlBulkCopyOptions::TableLock,
+                            &[],
+                        )
+                        .await?;
+                    bulk_insert_batch(&mut blk, &b, &collist).await?;
+                    blk.finalize().await?;
+                }
+                Err(l) => println!("{:?}", l),
+            },
+            None => break,
+        };
+    }
+    Ok(schema)
+}
